from typing import Any, Annotated, Literal

from accelforge.frontend.mapper.metrics import Metrics
from accelforge.util._basetypes import EvalableModel


class FFM(EvalableModel):
    """Configuration for the Fast and Fusiest Mapper."""

    metrics: Metrics = Metrics.ENERGY
    """ Metrics used to optimize mappings. """

    info_metrics: Metrics = Metrics.all_metrics()
    """Metrics to be reported for final mappings."""

    _timeloop_style_even: bool = False
    """ Timeloop-style even mappings must have, for each memory, at most two
        locations where storage nodes may be placed. """

    force_memory_hierarchy_order: bool = True
    """
    If set to true, storage nodes for lower-level memories must be placed below storage
    nodes for higher-level memories. For example, all MainMemory storage nodes must go
    above all LocalBuffer storage nodes.

    This constraint always applies to same-tensor storage nodes (e.g., MainMemory
    reusing Output must go above LocalBuffer reusing Output); turning it off will
    permit things like MainMemory reusing Output going above LocalBuffer reusing
    Input.
    """

    out_of_order_hierarchy_explore_removing_spatials_for_more_temporals: bool = False
    """
    If force_memory_hierarchy_order is set to False or is set to False for any
    particular component, and a spatial fanout ends up being raised above a storage node
    that does not have that fanout, then there may be cases where a spatial loop is put
    above a component that does not have the associated fanout.

    When this happens, we may not put between the spatial and the storage node any
    temporal loops that affect the same indexing expressions as the spatial loops.

    For example, the following is not allowed:

    Arch:

    - Global Buffer
    - 2x fanout
    - Register

    Mapping:

    spatial-for-reg n in [0, 10):
      [Register reuses input]
        for n in [0, 2):
          [Global Buffer reuses output]

    By default, if there are spatial loops that are not constrained away, then the
    mapper will not explore putting any temporal loops that conflict. In the above
    example, it will never place the above temporal loop. If this is set to True, then
    the mapper will explore removing the spatial loop in order to allow for the temporal
    loop to be placed. In the above example, it will explore removing the spatial loop
    in order to allow for the temporal loop to be placed.
    """

    max_fused_loops_per_rank_variable: int = 1
    """ The maximum number of fused loops in a pmapping for a given rank variable. """

    max_fused_loops: float | int = float("inf")
    """ The maximum total number of fused loops in a pmapping. """

    max_loops: float | int = float("inf")
    """ The maximum total loops in a pmapping. """

    max_loops_minus_ranks: float | int = float("inf")
    """
    The maximum total loops in a pmapping minus the number of ranks. For example,
    3 means that the number of loops can be up to (the number of ranks + 3).
    """

    _can_lower_outermost_memory: bool = False
    """
    Whether the storage node of outermost memory can be lowered. If set to True, the
    mapper may exchange tiles of tensors via the outermost memory, instead of storing
    full tensors. Set this to True to explore reducing outermost memory usage.

    TODO: Also need to explore putting loops above the outermost memory then. This is
    currently private because we may want to have a catch-all term like
    "save_outermost_memory_usage".
    """

    _only_output_pmapping_with_index: int | None = None
    """
    For debugging. Only output the pmapping with this index.
    """

    memory_limit: float | int = float("inf")
    """ The maximum memory limit for the mapper. """

    memory_limit_per_process: float | int = float("inf")
    """ The maximum memory limit per process for one of the mapper's processes. """

    time_limit: float | int = float("inf")
    """ The maximum time limit for the mapper. """

    time_limit_per_pmapping_template: float | int = float("inf")
    """ The maximum time limit per pmapping template. """

    max_pmapping_templates_per_einsum: float | int = float("inf")
    """
    The maximum number of pmapping templates per Einsum. Once this many templates are
    generated, the mapper will stop generating more. This is useful for debugging (why
    are so many templates being generated?).
    """

    _count_option_for_mapsapce_size_evaluation: tuple[
        Literal[
            "redundant_loop_orders",
            "non_helpful_loops_for_loop_orders",
            "non_helpful_tile_shapes",
            "redundant_dataplacements",
        ]
    ] = ()



import unittest

from accelforge.frontend.spec import Spec
from accelforge.model.main import evaluate_mapping
from accelforge.util.parallel import set_n_parallel_jobs
from accelforge.plotting.mappings import (
    plot_energy_comparison,
    plot_energy_breakdown,
    plot_action_breakdown,
)

set_n_parallel_jobs(1)

try:
    from .paths import EXAMPLES_DIR
except ImportError:
    from paths import EXAMPLES_DIR


class TestEnergyPlotting(unittest.TestCase):
    def test_comparison(self):
        spec = Spec.from_yaml(
            EXAMPLES_DIR / "arches" / "simple.yaml",
            EXAMPLES_DIR / "workloads" / "matmuls.yaml",
            EXAMPLES_DIR / "mappings" / "unfused_matmuls_to_simple.yaml",
            jinja_parse_data={"N_EINSUMS": 2, "M": 64, "KN": 64},
        )

        result = evaluate_mapping(spec)

        fig, ax = plot_energy_comparison([result])
        fig.tight_layout()
        fig.savefig("total.png", dpi=400, bbox_inches="tight")

    def test_breakdown(self):
        spec = Spec.from_yaml(
            EXAMPLES_DIR / "arches" / "simple.yaml",
            EXAMPLES_DIR / "workloads" / "matmuls.yaml",
            EXAMPLES_DIR / "mappings" / "unfused_matmuls_to_simple.yaml",
            jinja_parse_data={"N_EINSUMS": 2, "M": 64, "KN": 64},
        )

        result = evaluate_mapping(spec)

        fig, axes = plot_energy_breakdown([result, result], ["einsum", "component"])
        fig.tight_layout()
        fig.savefig("/tmp/fig.png", dpi=400, bbox_inches="tight")

    def test_breakdown_stacked(self):
        spec = Spec.from_yaml(
            EXAMPLES_DIR / "arches" / "simple.yaml",
            EXAMPLES_DIR / "workloads" / "matmuls.yaml",
            EXAMPLES_DIR / "mappings" / "unfused_matmuls_to_simple.yaml",
            jinja_parse_data={"N_EINSUMS": 2, "M": 64, "KN": 64},
        )

        result = evaluate_mapping(spec)

        fig, axes = plot_energy_breakdown(
            [result, result], ["einsum", "component"], ["action"]
        )
        fig.tight_layout()
        fig.savefig("/tmp/fig.png", dpi=400, bbox_inches="tight")


class TestActionPlotting(unittest.TestCase):
    def test_breakdown(self):
        spec = Spec.from_yaml(
            EXAMPLES_DIR / "arches" / "simple.yaml",
            EXAMPLES_DIR / "workloads" / "matmuls.yaml",
            EXAMPLES_DIR / "mappings" / "unfused_matmuls_to_simple.yaml",
            jinja_parse_data={"N_EINSUMS": 2, "M": 64, "KN": 64},
        )

        result = evaluate_mapping(spec)

        fig, axes = plot_action_breakdown([result, result], ["einsum", "component"])
        fig.tight_layout()
        fig.savefig("/tmp/fig.png", dpi=400, bbox_inches="tight")



from accelforge.frontend.mapping.mapping import MappingNode


import copy
from collections import defaultdict
import itertools
import logging
from typing import Any, Iterator, List
import uuid

from tqdm import tqdm

import accelforge.frontend.arch as arch
from accelforge.frontend.mapping import (
    Compute,
    Loop,
    Mapping,
    MappingNode,
    Spatial,
    TensorHolder,
    Temporal,
)
from accelforge.frontend.spec import Spec
from accelforge.frontend._workload_isl._isl import get_rank_variable_bounds
from accelforge.frontend._workload_isl._symbolic import (
    Relevant,
    get_rank_variable_relevancy,
    get_stride_and_halo,
    get_stride_and_halo_of_einsum,
    PartiallyRelevant,
)
from accelforge.frontend.workload import (
    TensorName,
    Einsum,
    EinsumName,
    RankVariable,
    Workload,
    isl_expression_has_variable,
    SymbolTable,
)
from accelforge.mapper.FFM._make_pmappings.make_pmapping_templates.make_storage_order import (
    get_tensor_choices,
)
from accelforge.mapper.FFM._make_pmappings.make_pmapping_templates.make_reservations import (
    get_reservation_choices,
)
from accelforge.mapper.FFM._make_pmappings.contraints.constraints import (
    MappingConstraints,
    get_constraints,
)
from accelforge.mapper.FFM._make_pmappings.make_pmapping_templates.make_loops import (
    insert_temporal_loops,
    insert_spatial_loops,
)
from accelforge.mapper.FFM._make_pmappings.pmapper_job import (
    Job,
    SameEinsumJobs,
)
from accelforge.model._looptree.reuse.symbolic import label_fused_loops


def unpack_loops_to_rank_variables(mapping: List[MappingNode]):
    mapping_new = []
    for node in mapping:
        if not isinstance(node, Loop) or not isinstance(node.rank_variable, set):
            mapping_new.append(node)
            continue

        for r in sorted(node.rank_variable):
            mapping_new.append(
                type(node)(
                    rank_variable=r,
                    **node.model_dump(exclude={"rank_variable"}, recursive=False),
                )
            )
    return mapping_new


# =================================================================================================
# Iterate over mappings
# =================================================================================================
def place_missing_temporal_loops(
    mapping: List[MappingNode], einsum: Einsum, flattened_arch: list[arch.Leaf]
):
    """
    Adds temporal loops to the mapping to fill in any rank variables that are missing.
    This may occur if there are no points where it'd be helpful to add a non-fused loop,
    so we just need to add one somewhere.
    """
    # If any rank variables are missing, add them as high as possible.

    rank_variables = einsum.rank_variables
    for m in mapping:
        if isinstance(m, Temporal) and not m._fused:
            rank_variables.discard(m.rank_variable)

    # Insert point: Right under the last backing & below any out-of-order fanouts
    fanouts = {}
    fanout = 1
    for node in flattened_arch:
        fanouts[node.name] = (fanout := fanout * node.get_fanout())

    insert_point = 0
    greatest_previous_fanout = 1
    for i in range(len(mapping)):
        if isinstance(mapping[i], TensorHolder):
            if mapping[i]._backing:
                insert_point = i + 1
            cur_fanout = fanouts[mapping[i].component]
            if cur_fanout < greatest_previous_fanout:
                insert_point = i + 1
            greatest_previous_fanout = max(greatest_previous_fanout, cur_fanout)

        # Put it below all the other temporals here in case we're lowering through them
        if isinstance(mapping[i], Temporal) and insert_point == i:
            insert_point = i + 1

    temporals = [Temporal(rank_variable=r) for r in sorted(rank_variables)]

    if insert_point == len(mapping):
        mapping.extend(temporals)
    else:
        for t in temporals:
            mapping.insert(insert_point, t)


def remove_unordered_spatial_temporal_loops(
    mapping: list[MappingNode],
    flattened_arch: list[arch.Leaf],
    einsum: Einsum,
    explore_unordered_spatial_loops: bool = True,
):
    fanout = 1
    fanouts = {}
    for node in flattened_arch:
        fanouts[node.name] = (fanout := fanout * node.get_fanout())

    index_exprs = einsum.indexing_expressions

    # Remove a temporal loop if:
    # - It's between a spatial loop and a storage node above that fanout in the arch
    # - It indexes into one of the same indexing expressions as the spatial loop

    disallowed_combinations: list[tuple[set[int], set[int]]] = []
    for i, node in enumerate(mapping):
        if not isinstance(node, Spatial):
            continue

        last_idx_to_check = _idx_of_lowest_tensor_holder_with_component_above_fanout(
            mapping, i, fanouts, node
        )
        to_check = mapping[i + 1 : last_idx_to_check]
        to_remove = set()
        for n in to_check:
            if isinstance(n, Temporal):
                for expr in index_exprs:
                    if not isl_expression_has_variable(expr, node.rank_variable):
                        continue
                    if not isl_expression_has_variable(expr, n.rank_variable):
                        continue
                    to_remove.add(id(n))
                    break

        if to_remove:
            disallowed_combinations.append((set([id(node)]), to_remove))

    if not explore_unordered_spatial_loops:
        disallowed_combinations = [x[1:] for x in disallowed_combinations]

    for combo in itertools.product(*disallowed_combinations):
        combo = set.union(set(), *combo)
        yield [n for n in mapping if id(n) not in combo]


def _idx_of_lowest_tensor_holder_with_component_above_fanout(
    mapping, start_idx, fanouts, node
):
    """
    Return idx of lowest tensor holder with component above fanout. If none
    found, returns index right under start idx (start_idx + 1).
    """
    for j in range(len(mapping) - 1, start_idx, -1):
        n = mapping[j]
        if (
            isinstance(n, TensorHolder)
            and fanouts[n.component] < fanouts[node.component]
        ):
            return j
    return start_idx + 1


def pad_with_bottom_loops(mapping: list[MappingNode], einsum: Einsum):
    rank_variables = einsum.rank_variables
    rank_var_to_count = defaultdict(lambda: 0)
    for node in mapping:
        if isinstance(node, Temporal):
            rank_var_to_count[node.rank_variable] += 1

    for rank_var in rank_variables:
        if rank_var_to_count[rank_var] < 2:
            mapping.append(Temporal(rank_variable=rank_var))


def _timeloop_style_even(mapping: list[MappingNode]):
    # Iterate through the mapping. If there are >2 TensorHolder nodes for the same
    # memory, move all below the 2nd to the same level as the 2nd.
    mapping = copy.deepcopy(mapping)
    memory2indices = defaultdict(list)
    i = 0
    while i < len(mapping):
        node = mapping[i]
        if not isinstance(mapping[i], TensorHolder):
            i += 1
            continue
        node: TensorHolder
        seen = memory2indices[node.component]
        mapping[i]._lower = False  # Lowering might re-uneven the reservationsxs

        if len(seen) <= 1:
            seen.append(i)
        else:
            mapping.insert(seen[-1] + 1, mapping.pop(i))
        i += 1
    return mapping


def assert_proper_fusion_labeling(
    mapping: list[MappingNode],
    fusable_tensors: set[TensorName],
    check_loops: bool = True,
):
    tensors = set()
    for i, t in enumerate(mapping):
        if not isinstance(t, TensorHolder):
            continue

        new = (set(t.tensors) - tensors) & fusable_tensors

        if new and check_loops:
            for j in range(i):
                if isinstance(mapping[j], Loop):
                    assert mapping[
                        j
                    ]._fused, f"Node {j} is not fused in {' '.join(m.compact_str() for m in mapping)}"
        assert (
            t._backing & fusable_tensors
        ) == new, f"Node {i} backing missing {new - t._backing} in {' '.join(m.compact_str() for m in mapping)}"
        tensors.update(new)
        tensors.update(t.tensors)


def get_initial_delta_choices(einsum_name: str, workload: Workload):
    stride_and_halo = get_stride_and_halo(workload)
    einsum = workload.einsums[einsum_name]

    choices = defaultdict(lambda: set([0]))
    consumer_chains = []
    stack = [[(None, einsum)]]
    while stack:
        cur_chain = stack.pop()
        last_tensor, last_einsum = cur_chain[-1]
        for tensor in last_einsum.output_tensor_names:
            einsums_with_tensor_as_input = workload.einsums_with_tensor_as_input(tensor)

            if len(einsums_with_tensor_as_input) == 0:
                consumer_chains.append(cur_chain)

            for next_einsum in einsums_with_tensor_as_input:
                stack.append(cur_chain + [(tensor, next_einsum)])

    for chain in consumer_chains:
        for (_, producer), (tensor, consumer) in zip(
            list(reversed(chain))[1:], reversed(chain)
        ):
            rank_stride_and_halo = stride_and_halo[(consumer.name, tensor)]
            if tensor is None:
                break  # done

            for cons_rank_var in consumer.rank_variables:
                for prod_rank_var in producer.rank_variables:
                    for cons_choice in choices[cons_rank_var]:
                        if (prod_rank_var, cons_rank_var) not in rank_stride_and_halo:
                            continue
                        stride, halo = rank_stride_and_halo[
                            (prod_rank_var, cons_rank_var)
                        ]
                        choices[prod_rank_var].add(cons_choice * stride + halo)

    return choices


def get_ranks_with_tile_pattern(producer_name: EinsumName, workload: Workload):
    initial_choices = get_initial_delta_choices(producer_name, workload)
    return {
        rank_var
        for rank_var in workload.einsums[producer_name].rank_variables
        if len(initial_choices[rank_var]) > 1
    }


def iterate_mappings_no_constraints(
    spec: Spec,
    einsum_name: str,
    flattened_arch: list[arch.Leaf],
    rank_variable_bounds: dict[RankVariable, int],
    job: Job,
) -> Iterator[tuple[Mapping, SymbolTable, arch.Compute, int]]:
    first_memory = None
    for node in flattened_arch:
        if isinstance(node, arch.Memory):
            first_memory = node
            break
    if first_memory is None:
        raise ValueError("No memory found in architecture")

    ranks_with_tile_pattern = get_ranks_with_tile_pattern(einsum_name, spec.workload)

    einsum = spec.workload.einsums[einsum_name]
    symbol_table = {r.name: r.source for r in einsum.renames}
    fusable_tensors = job.fusable_tensors

    fanouts = {}
    fanout = 1
    for node in flattened_arch:
        fanouts[node.name] = (fanout := fanout * node.get_fanout())

    for mapping, symbol_table, compute in get_tensor_choices(
        einsum_name,
        flattened_arch,
        symbol_table,
        spec,
        first_memory,
        fusable_tensors,
        fanouts,
    ):
        logging.info(
            "\tGenerated tensor choices: " + ", ".join(m.compact_str() for m in mapping)
        )
        mapping = copy.deepcopy(mapping)
        for mapping, n_orders in insert_temporal_loops(
            mapping,
            einsum,
            first_memory,
            rank_variable_bounds,
            ranks_with_tile_pattern,
            spec.workload,
            spec.mapper._can_lower_outermost_memory,
            flattened_arch,
            spec.mapper.max_fused_loops,
            fanouts,
        ):
            mapping = copy.deepcopy(mapping)
            insert_spatial_loops(mapping, einsum, flattened_arch)
            mapping = unpack_loops_to_rank_variables(mapping)
            if spec.mapper._timeloop_style_even:
                mapping = _timeloop_style_even(mapping)

            place_missing_temporal_loops(mapping, einsum, flattened_arch)
            label_fused_loops(mapping, fusable_tensors)
            assert_proper_fusion_labeling(mapping, fusable_tensors)
            yield mapping, symbol_table, compute, n_orders


def iterate_mappings_constraints(
    spec: Spec,
    einsum_names: list[str] | str,
    flattened_arch: list[arch.Leaf],
    rank_variable_bounds: dict[RankVariable, int],
    tensor_to_relevancy: dict[
        TensorName, dict[RankVariable, Relevant | PartiallyRelevant]
    ],
    job: Job,
) -> Iterator[tuple[Mapping, MappingConstraints, dict[str, str]]]:
    compute_name = flattened_arch[-1].name

    n_yielded = 0

    if isinstance(einsum_names, str):
        einsum_names = [einsum_names]

    for einsum_name in einsum_names:
        logging.info(
            f"Generating pmapping templates for compute {compute_name} Einsums "
            f"{einsum_name}"
        )

        for mapping, symbol_table, compute, n_orders in iterate_mappings_no_constraints(
            spec,
            einsum_name,
            flattened_arch,
            rank_variable_bounds,
            job,
        ):
            mapping, constraints = get_constraints(
                flattened_arch, mapping, symbol_table, einsum_name, tensor_to_relevancy
            )

            # This goes after the constraints because constraints may remove some loops,
            # giving us fewer that may be reordered.
            for mapping in remove_unordered_spatial_temporal_loops(
                mapping,
                flattened_arch,
                spec.workload.einsums[einsum_name],
                spec.mapper.out_of_order_hierarchy_explore_removing_spatials_for_more_temporals,
            ):
                constraints.remove_missing_targets(mapping)

                mapping.append(
                    Compute(
                        einsum=einsum_name,
                        component=compute_name,
                        component_object=compute,
                    )
                )

                # MAPPING MUST NOT BE MODIFIED AFTER constraints.set_loop_indices
                constraints.set_loop_indices(mapping)

                mapping = Mapping(nodes=[copy.copy(n) for n in mapping])
                mapping._n_loop_orders = n_orders
                yield mapping, constraints, symbol_table
                n_yielded += 1
                if n_yielded >= spec.mapper.max_pmapping_templates_per_einsum:
                    return


# =================================================================================================
# Top level
# =================================================================================================
def make_pmapping_templates(job: Job, print_progress: bool = True) -> SameEinsumJobs:
    compute_name = job.flattened_arch[-1].name

    job.tensor_to_relevancy = {
        tensor: get_rank_variable_relevancy(
            job.spec.workload.einsums[job.einsum_name], tensor
        )
        for tensor in job.spec.workload.einsums[job.einsum_name].tensor_names
    }

    mappings_constraints = iterate_mappings_constraints(
        job.spec,
        job.einsum_name,
        job.flattened_arch,
        job.rank_variable_bounds,
        job.tensor_to_relevancy,
        job,
    )
    if print_progress:
        mappings_constraints = tqdm(
            mappings_constraints,
            desc=f"Generating pmapping templates for compute {compute_name} Einsum {job.einsum_name}",
        )

    stride_and_halo = get_stride_and_halo_of_einsum(
        job.einsum_name, job.spec.workload, job.rank_variable_bounds
    )

    jobs = SameEinsumJobs()
    only_output_pmapping_index = job.spec.mapper._only_output_pmapping_with_index
    for i, (mapping, constraints, symbol_table) in enumerate(mappings_constraints):
        if only_output_pmapping_index is not None and i != only_output_pmapping_index:
            continue
        new_job = copy.copy(job)
        new_job.mapping = mapping
        new_job.constraints = constraints
        new_job.job_id = uuid.uuid4()
        new_job.rank_variable_bounds = job.rank_variable_bounds
        new_job.stride_and_halo = stride_and_halo
        new_job.compatibility
        jobs.append(new_job)

    return jobs



import accelforge as af
from accelforge.tracegen.tracemapping import trace_iterations


class TestUnfused:
    def test_unfused_matvecs_to_simple(self):
        spec = af.Spec.from_yaml(
            af.examples.arches.simple,
            af.examples.workloads.matvecs,
            af.examples.mappings.unfused_matvecs_to_simple,
            jinja_parse_data={"N_EINSUMS": 2, "KN": 4, "MainMemoryEnergy": 10}
        )
        result = spec.evaluate_mapping()
        trace_iterations(result.mapping(), spec)



from collections.abc import Generator
from typing import Any

import accelforge.frontend.arch as arch
from accelforge.frontend.mapping import MappingNode, Reservation, Storage, TensorHolder
from accelforge.frontend.spec import Spec


def _recursive_iter_fence_positions(
    fence_positions: dict[str, int],
    max_size: int,
) -> Generator[tuple[list[TensorHolder], Any], None, None]:
    if not fence_positions:
        yield {}
    mine = next(iter(fence_positions))
    myval = fence_positions[mine]
    following = {k: v for k, v in fence_positions.items() if k != mine}
    for i in range(myval, max_size):
        following = {k: max(v, i) for k, v in fence_positions.items() if k != mine}
        for following in _recursive_iter_fence_positions(following, max_size):
            yield {mine: i, **following}


def get_reservation_choices(
    mapping: list[TensorHolder],
    flattened_arch: list[arch.Leaf],
) -> Generator[tuple[list[TensorHolder], Any], None, None]:
    # Rules:
    # - In general, reservations go right under their storage node
    # - If a storage node is associated with a fanout, explore putting the reservation
    #   below it, below the next storage node, and so on. Stop once we don't have any
    #   more spatial loops to place. Push down all reservations below this fanout
    #   together.

    # Spatial loops:
    # - Must go below all storage nodes associated with something above the fanout.
    #   -> Memories above fanout must serve all fetches across fanout instances.
    # - Must go above all reservations associated with something below the fanout.
    #   -> Memories below fanout must be reserved for each fanout instance.
    # - If below any storage node associated with the fanout, then must be relevant.
    #   -> No peer-to-peer communication

    # Temporal loops:
    # - If between a storage node and a reservation node, the outermost temporal loop
    #   may be partially relevant. All others must be relevant.

    # Design choices here:
    # - Where to put the 'fence' for each fanout

    fanout_nodes = [n for n in flattened_arch if n.get_fanout() > 1]
    fanout_node_names = set[str](n.name for n in fanout_nodes)
    last_seen_fanout = None
    node2lastfanout = {}

    fence_positions: dict[str, int] = {}
    for i, node in enumerate(mapping):
        if node.component in fanout_node_names:
            fence_positions.setdefault(node.component, i)
            last_seen_fanout = node.component
        node2lastfanout[id(node)] = last_seen_fanout

    def try_add_reservations(
        new_mapping: list[MappingNode],
        reservations_to_add: list[TensorHolder],
        fence_positions: dict[str, int],
    ):
        for res in list(reservations_to_add):
            add = False
            if node2lastfanout[id(res)] is None:
                add = True
            elif i >= fence_positions[node2lastfanout[id(res)]]:
                add = True
            if add:
                new_mapping.append(
                    Reservation(
                        purposes=[res.component],
                        resource=res.component,
                        persistent=res.persistent,
                    )
                )
                reservations_to_add.remove(res)

    # Fence positions are indices of storage nodes below which we'll push all the
    # reservations below that fanout
    for fence_positions in _recursive_iter_fence_positions(
        fence_positions, len(mapping)
    ):
        new_mapping = []
        reservations_to_add = []
        for i, node in enumerate(mapping):
            new_mapping.append(node)
            reservations_to_add.append(node)
            try_add_reservations(new_mapping, reservations_to_add, fence_positions)
        try_add_reservations(new_mapping, reservations_to_add, fence_positions)
        yield new_mapping



"""
Some non-exhaustive tests to make sure core functionality of the non-fusing and
fusing portions of ISL tiling is not violated under changes.
"""

from multiprocessing.sharedctypes import Value
from islpy import Val
from pathlib import Path
from pprint import pformat
import unittest

import islpy as isl

from accelforge.frontend.workload import Workload
from accelforge.frontend.mapping import Mapping

from accelforge.model._looptree.reuse.isl.mapping_to_isl import analyze_mapping
from accelforge.model._looptree.reuse.isl.mapping_to_isl.types import (
    MappingAnalysisResult,
)

from ..util import load_solutions

TEST_CONFIG_PATH: Path = Path(__file__).parent / "configs"


class TestMappingToIsl(unittest.TestCase):
    """
    Tests taking a `Mapping` and `Workload` and converting it into relevant isl
    objects.
    """

    def test_conv1d(self):
        """
        Non-fusing one 1-dimensional convolution test.
        """
        # Loads in the CONV1D Config
        config_path: Path = TEST_CONFIG_PATH / "conv1d"
        workload: Workload = Workload.from_yaml(
            config_path / "conv1d.workload.yaml", top_key="workload"
        )

        mapping: Mapping = Mapping.from_yaml(
            config_path / "conv1d.mapping.yaml", top_key="mapping"
        )
        occupancies: MappingAnalysisResult = analyze_mapping.occupancies_from_mapping(
            mapping, workload
        )

        for buffer, occupancy in occupancies.buffet_to_occupancy.items():
            if buffer == list(occupancies.buffet_to_occupancy.keys())[-1]:
                soln: isl.Map = isl.Map.read_from_str(
                    isl.DEFAULT_CONTEXT,
                    "{ MAC_spacetime[P1, P0, R] -> O[P=8*P1 + P0] : "
                    "0 <= R < 3 and 0 <= P1 < 2 and 0 <= P0 < 8}",
                )
                assert occupancy.map_ == soln

    def test_two_conv1d(self):
        """
        Fusing two 1-dimensional convolutions test.
        """
        # Loads in the CONV1D Config
        config_path: Path = TEST_CONFIG_PATH / "two_conv1d"
        workload: Workload = Workload.from_yaml(
            config_path / "two_conv1d.workload.yaml", top_key="workload"
        )

        mapping: Mapping = Mapping.from_yaml(
            config_path / "two_conv1d.mapping.yaml", top_key="mapping"
        )
        occupancies: MappingAnalysisResult = analyze_mapping.occupancies_from_mapping(
            mapping, workload
        )
        solns: dict = load_solutions(config_path / "two_conv1d.expected.yaml")[
            "mapping_to_isl"
        ]

        errors: list = []
        buffers_seen: set = set()
        for buffer, occupancy in occupancies.buffet_to_occupancy.items():
            try:
                soln = solns[repr(buffer)]
                assert (
                    occupancy.map_ == soln
                ), (
                    f"{buffer} should hold:\n" +
                    f"{soln}\n" +
                    f"instead holds:\n"
                    f"{occupancy.map_}\n" +
                    '-'*3
                )
            except (AssertionError, KeyError) as e:
                errors.append(e)
            buffers_seen.add(repr(buffer))

        buffers_known: set = set(solns.keys())
        try:
            assert buffers_seen == buffers_known, (
                f"Buffers Missing: {pformat(buffers_seen - buffers_known)}\n"
                f"Buffers Extra: {pformat(buffers_known - buffers_seen)}"
            )
        except AssertionError as e:
            errors.append(e)

        if len(errors) != 0:
            for e in errors:
                print('#' * 15)
                print(e)
            raise ValueError("There were errors in the two_conv1d results (see logs)")



from collections import defaultdict
from enum import Enum
from dataclasses import dataclass, replace
import itertools
from numbers import Number
from typing import Literal, TypeVar

import pandas as pd
from accelforge.frontend.mapping import (
    Loop,
    Mapping,
    Spatial,
    TensorHolder,
    Reservation as MappingReservation,
    Split as MappingSplit,
    TilePattern,
    Loop as MappingLoop,
)
from accelforge.frontend.renames import Rank, RankVariable, TensorName
from accelforge.mapper.FFM._pareto_df.df_convention import (
    is_fused_loop_col,
    make_fused_loop_col,
    stride2col,
    initial2col,
    iterations2col,
)

from accelforge.util import _expfmt, fzs

# Abstractions:
# 1. Each tensor is stored above some loop index. 0 is the outermost loop, 1 the
#    next-innermost...
# 2. All loops above any shared tensor are co-tiled and must match between PmappingGroups.

T = TypeVar("T", bound="Updatable")


class Updatable:
    def update(self: T, **kwargs) -> T:
        return replace(self, **kwargs)


def _update_rename_dict(
    renames: dict[str, str],
    new_renames: dict[str, str],
):
    for mine, other in new_renames.items():
        if mine not in renames:
            renames[mine] = other
        elif renames[mine] != other:
            raise ValueError(
                f"Renaming {mine} to {other} conflicts with {renames[mine]}"
            )


@dataclass(frozen=True, order=True, eq=True)
class Loop(Updatable):
    rank_name: Rank
    tile_pattern: TilePattern | None
    is_spatial: bool

    def __post_init__(self):
        assert isinstance(self.rank_name, Rank)
        assert isinstance(self.tile_pattern, Number | TilePattern | str | None)
        assert isinstance(self.is_spatial, bool)
        assert isinstance(
            self.tile_pattern.initial_tile_shape,
            Number | str | None,
        )
        assert isinstance(
            self.tile_pattern.tile_shape,
            Number | str | None,
        )

    def __repr__(self):
        return (
            f"Loop({self.rank_name.__repr__()}, {self.tile_pattern}, {self.is_spatial})"
        )

    def __str__(self):
        return (
            "S-" if self.is_spatial else ""
        ) + f"{self.rank_name}-{self.tile_pattern}"

    def pydot_str(self):
        if self.is_spatial:
            return f"S-for R{self.rank_name} size {_expfmt(self.tile_pattern)}"
        return f"for {self.rank_name} size {_expfmt(self.tile_pattern)}"

    def to_yaml(self):
        return {"type": "loop", **self.__dict__}

    def merge_next(self, right: "Loop") -> "Loop":
        assert self.tile_pattern == right.tile_pattern
        return Loop(
            self.rank_name | right.rank_name,
            right.tile_pattern,
            self.is_spatial,
        )

    def clear_loop_bound(self, value=0):
        return self.update(tile_pattern=value)

    def populate(self, nloop: int) -> "Loop":
        tile_pattern = TilePattern(
            tile_shape=stride2col(self.rank_name, nloop),
            initial_tile_shape=initial2col(self.rank_name, nloop),
            calculated_n_iterations=iterations2col(nloop),
        )
        return self.update(tile_pattern=tile_pattern)

    def _prepend_symbols(self, prepend: str) -> "Loop":
        return self.update(tile_pattern=self.tile_pattern._prepend_symbols(prepend))

    def clear_symbolic_tile_patterns(self) -> "Loop":
        return self.update(tile_pattern=self.tile_pattern._clear_symbols())

    def make_fused_loop_symbols(self, prefix: str) -> tuple[dict[str, str], "Loop"]:
        r = {}
        new = self

        def replace(attr, new):
            g = getattr(self.tile_pattern, attr)
            if not isinstance(g, str):
                return new
            g2 = make_fused_loop_col(f"{prefix}<SEP>{g}")
            r[g] = g2
            return new.update(tile_pattern=new.tile_pattern.update(**{attr: g2}))

        for s in new.tile_pattern._symbol_attrs():
            new = replace(s, new)

        return r, new

    def add_n_iteration_symbols(self) -> "Loop":
        return self.update(tile_pattern=self.tile_pattern.add_n_iteration_symbols())

    def _rename_to_match(self, other: "Loop") -> tuple["Loop", dict[str, str]]:
        new_tp, renames = self.tile_pattern._rename_to_match(other.tile_pattern)
        return self.update(rank_name=other.rank_name, tile_pattern=new_tp), renames


@dataclass(frozen=True, eq=True, order=True)
class TensorReservation(Updatable):
    # This order is important. Above loop index should be before resource name
    # so when we sort reservations for tensors then the backing tensor holder comes
    # first.
    # Size is not included in hash or equality functions. This is because there
    # may be floating point rounding errors in reservation sizes. The other
    # attributes are sufficient to determine equality.
    loops: tuple[Loop]
    name: TensorName
    resource_name: str
    persistent: bool = False

    def __post_init__(self):
        if self.persistent:
            assert len(self.loops) == 0, "Persistent tensors be above all loops"

    @property
    def above_loop_index(self) -> int:
        return -1 if self.persistent else len(self.loops)

    def __str__(self):
        return f"[{self.resource_name}] {self.name} below {self.loops}"

    def __repr__(self):
        return f"Reservation({repr(self.name)}, {repr(self.loops)}, {repr(self.resource_name)})"

    def pydot_str(self):
        return f"[{self.resource_name}] {self.name}"

    def permute(self, permutation) -> "Reservation":
        new_loops = [self.loops[permutation[i]] for i in range(len(self.loops))]
        return self.update(loops=tuple(new_loops))

    def clear_loop_bounds(self) -> "Reservation":
        return self.update(loops=tuple(loop.clear_loop_bound() for loop in self.loops))

    def populate_loops(self) -> "TensorReservation":
        return self.update(
            loops=tuple(loop.populate(nloop) for nloop, loop in enumerate(self.loops))
        )

    @staticmethod
    def get_backing_tensors(
        all_tensors: set["TensorReservation"],
    ) -> list["TensorReservation"]:
        id2tensor = defaultdict(lambda: [])
        for t in all_tensors:
            id2tensor[t.name].append(t)
        return sorted(sorted(v)[0] for v in id2tensor.values())

    def drop_loop_indices(self, loop_indices: set[int]) -> "TensorReservation":
        loops = tuple(l for i, l in enumerate(self.loops) if i not in loop_indices)
        return self.update(loops=loops)

    def _prepend_symbols(self, prepend: str) -> "TensorReservation":
        return self.update(
            loops=tuple(l._prepend_symbols(prepend) for l in self.loops),
        )

    def clear_symbolic_tile_patterns(self) -> "TensorReservation":
        return self.update(
            loops=tuple(l.clear_symbolic_tile_patterns() for l in self.loops),
        )

    def make_fused_loop_symbols(
        self, prefix: str
    ) -> tuple[dict[str, str], "TensorReservation"]:
        result = {}
        loops = []
        for l in self.loops:
            r, l = l.make_fused_loop_symbols(prefix)
            result.update(r)
            loops.append(l)
        return result, self.update(loops=tuple(loops))

    def add_n_iteration_symbols(self) -> "TensorReservation":
        return self.update(
            loops=tuple(l.add_n_iteration_symbols() for l in self.loops),
        )

    def _rename_to_match(
        self, other: "TensorReservation"
    ) -> tuple["TensorReservation", dict[str, str]]:
        renames = {}
        new_loops = []
        for l_mine, l_other in zip(self.loops, other.loops):
            l_mine, new_renames = l_mine._rename_to_match(l_other)
            _update_rename_dict(renames, new_renames)
            new_loops.append(l_mine)
        return self.update(loops=tuple(new_loops)), renames


class SplitKind(Enum):
    SEQUENTIAL = 0
    PIPELINE = 1


@dataclass(frozen=True, order=True, eq=True)
class Split:
    split: MappingSplit
    above_loop_index: int


@dataclass(frozen=True)
class Compatibility(Updatable):
    tensors: fzs[TensorReservation]
    splits: fzs[Split] = fzs()
    reservation_indices: fzs[int] = fzs()
    check_reservation_indices: bool = True

    @property
    def n_loops(self) -> int:
        return max([len(s.loops) for s in self.tensors], default=0)

    @property
    def loops(self) -> tuple[Loop, ...]:
        return max([t.loops for t in self.tensors], key=len) if self.tensors else ()

    def _get_hash_tuple(self):
        return self.n_loops, self.tensors, self.reservation_indices

    def __hash__(self):
        return hash(self._get_hash_tuple())

    def __eq__(self, other):
        return self._get_hash_tuple() == other._get_hash_tuple()

    def __post_init__(self):
        assert isinstance(self.n_loops, int)
        assert isinstance(self.tensors, fzs)
        assert isinstance(self.splits, fzs)
        assert isinstance(self.reservation_indices, fzs)
        assert (
            max(self.reservation_indices, default=-1) <= self.n_loops
        ), f"Extra reservation indices {self.reservation_indices} are greater than n_loops {self.n_loops}"
        if self.check_reservation_indices:
            p = f"are not in reservation indices {self.reservation_indices}"
            assert all(
                i >= 0 for i in self.reservation_indices
            ), f"Reservation indices {self.reservation_indices} are not all >= 0"
            assert all(
                s.above_loop_index in self.reservation_indices for s in self.splits
            ), f"Split above loop indices {self.splits} {p}"
            assert all(
                len(s.loops) in self.reservation_indices for s in self.tensors
            ), f"Tensor loops {self.tensors} {p}"

    def get_backing_levels(self) -> dict[str, int]:
        backings = {}
        for t in self.tensors:
            prev = backings.get(t.name, t.above_loop_index)
            backings[t.name] = min(prev, t.above_loop_index)
        return backings

    @property
    def tensor_names(self) -> set[str]:
        return {t.name for t in self.tensors}

    @property
    def max_above_loop_index(self) -> int:
        if len(self.tensors) == 0:
            return 0
        return max(s.above_loop_index for s in self.tensors)

    def shared_loop_index(self, live_tensors: set[str]) -> int:
        n = [l for t, l in self.get_backing_levels().items() if t in live_tensors]
        return max(n) - 1 if n else -1

    def __len__(self) -> int:
        return self.max_above_loop_index

    def _rename_to_match(
        self, other: "Compatibility"
    ) -> tuple["Compatibility", dict[str, str]]:
        renames = {}
        assert (
            self.clear_symbolic_tile_patterns() == other.clear_symbolic_tile_patterns()
        )
        tensors = []
        for t in self.tensors:
            other_t = other.get_tensor_by_name(t.name)
            t, new_renames = t._rename_to_match(other_t)
            tensors.append(t)
            _update_rename_dict(renames, new_renames)

        return (
            Compatibility(
                tensors=fzs(tensors),
                splits=self.splits,
                reservation_indices=self.reservation_indices,
            ),
            renames,
        )

    def clear_dead_tensors(
        self,
        live_tensors: set[str] | Literal["All"],
    ) -> "Compatibility":
        """
        Return a new compatibility with "dead" tensors removed by:
        1. keeping only loops relevant to `live_tensors` and
        2. keeping only `live_tensors`.

        If `keep_loops` is `True`, then all loops are kept.
        If `keep_tensors` is a set, tensors in the set are kept.
        """
        if live_tensors == "All":
            live_tensors = self.tensor_names

        remaining_tensors = fzs(s for s in self.tensors if s.name in live_tensors)
        new_n_loops = max((len(s.loops) for s in remaining_tensors), default=0)
        new_splits = fzs(
            split for split in self.splits if split.above_loop_index < new_n_loops
        )
        reservation_indices = fzs(
            {min(i, new_n_loops) for i in self.reservation_indices}
        )
        reservation_indices = fzs(x for x in reservation_indices if x >= 0)

        return self.update(
            tensors=remaining_tensors,
            splits=new_splits,
            reservation_indices=reservation_indices,
        )

    def __lt__(self, other):
        return self._get_hash_tuple() < other._get_hash_tuple()

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f"Compatibility(n_loops={self.n_loops}, tensors={repr(self.tensors)}), splits={repr(self.splits)}"

    def _and_tensors_with_names(self, names: set[str]) -> "Compatibility":
        return fzs(s for s in self.tensors if s.name in names)

    def merge_next(
        self,
        right: "Compatibility",
        live_tensors: set[str],
    ) -> "Compatibility":
        self_freed = self.clear_dead_tensors(live_tensors)
        right_freed = right.clear_dead_tensors(live_tensors)
        if self_freed.n_loops > right_freed.n_loops:
            # This can be relaxed if we have a way to do order-independent joining
            # and/or non-looptree mappings.
            raise ValueError(
                f"Can't merge. I have more loops than the next, so my dataflow can't "
                f"be carried through a LoopTree to where it's needed."
            )

        live_minus_mine = live_tensors - {s.name for s in self.tensors}
        tensors_a = self._and_tensors_with_names(live_tensors)
        tensors_b = right._and_tensors_with_names(live_minus_mine)

        # TODO: split handling?
        joined = Compatibility(
            tensors=tensors_a | tensors_b,
            reservation_indices=self_freed.reservation_indices
            | right_freed.reservation_indices,
        )

        return joined

    def has_tensor(self, *tensors: TensorReservation) -> bool:
        return all(any(s == t for s in self.tensors) for t in tensors)

    def _permute_stops(self):
        stops = set(len(s.loops) for s in self.tensors)
        stops |= self.reservation_indices
        stops |= set(s.above_loop_index for s in self.splits)
        return stops

    def permute(
        self,
        loop_changes: list[int],
    ) -> "Compatibility":
        assert len(loop_changes) <= self.n_loops
        assert set(loop_changes) == set(
            range(len(loop_changes))
        ), f"Loop changes {loop_changes} are not a permutation of {range(len(loop_changes))}"
        if len(loop_changes) < len(self.loops):
            loop_changes = loop_changes + list(
                range(len(loop_changes), len(self.loops))
            )

        permute_stops = self._permute_stops()
        for i, c in enumerate(loop_changes):
            for r in permute_stops:
                assert (i < r) == (
                    c < r
                ), f"Loop changes {loop_changes} cross reservation {r}"
        new_tensors = fzs(s.permute(loop_changes) for s in self.tensors)
        return self.update(tensors=new_tensors)

    def make_equivalent_permutations(self) -> list[tuple["Compatibility", list[int]]]:
        # Get contiguous blocks of loops with no tensor reservation between them
        blocks = []
        current_block = []
        permute_stops = self._permute_stops()
        for i in range(self.n_loops):
            # Can't permute loops if there's a reservation between them
            if i in permute_stops:
                blocks.append(current_block)
                current_block = []
            current_block.append(i)
        if current_block:
            blocks.append(current_block)

        per_block_permutations = [
            list(itertools.permutations(block)) for block in blocks
        ]
        all_permutations = list(itertools.product(*per_block_permutations))
        all_permutations = [
            list(itertools.chain(*loop_changes)) for loop_changes in all_permutations
        ]
        return [(self.permute(p), p) for p in all_permutations]

    def get_tensor_by_name(self, tensor: str) -> TensorReservation:
        for s in self.tensors:
            if s.name == tensor:
                return s
        raise ValueError(f"No reservation found for {tensor}")

    def per_tensor_compatibility(self) -> dict[str, "Compatibility"]:
        result = {}
        for s in self.tensors:
            result[s.name] = self.clear_dead_tensors(set([s.name]))
        return result

    def clear_loop_bounds(self) -> "Compatibility":
        return self.update(tensors=fzs(t.clear_loop_bounds() for t in self.tensors))

    def compatible_with(self, other: "Compatibility") -> bool:
        return True
        # for a in self.tensors:
        #     a = a.loops
        #     for b in other.tensors:
        #         b = b.loops
        #         if a[:len(b)] != b[:len(a)]:
        #             return False
        # return True

    def populate_loops(self):
        return self.update(
            tensors=fzs(t.populate_loops() for t in self.tensors),
        )

    @classmethod
    def from_mapping(
        cls,
        mapping: Mapping,
        tensors: set[TensorName],
        rank_variable_to_ranks: dict[TensorName, dict[RankVariable, Rank]],
    ) -> "Compatibility":
        # TODO: update compatibility to handle spatial-for loop per-tensor update
        tensor_indices = []
        split_above_loop_indices = []
        reservation_indices = []
        backing_remaining = set(tensors)
        n_seen_loops = 0
        n_fused_loops = 0
        for i, n in enumerate(mapping.nodes):
            if isinstance(n, MappingReservation):
                reservation_indices.append(n_seen_loops)
                if not (backing := set(n.purposes) & backing_remaining):
                    continue
                backing_remaining -= backing
                assert (
                    len(n.purposes) == 1
                ), "Backing reservations should have only one purpose"
                tensor_indices.append(i)
            elif isinstance(n, MappingSplit):
                split_above_loop_indices.append(n_seen_loops)
            elif isinstance(n, MappingLoop):
                n_seen_loops += 1
                n_fused_loops += bool(backing_remaining)
            elif isinstance(n, TensorHolder):
                reservation_indices.append(n_seen_loops)

        reservation_indices = fzs(min(n, n_fused_loops) for n in reservation_indices)
        reservation_indices = fzs(x for x in reservation_indices if x >= 0)

        assert (
            not backing_remaining
        ), f"Tensors {backing_remaining} not found in mapping"

        def get_rank(rank_variable, tensor):
            rv = rank_variable_to_ranks[tensor].get(rank_variable, set())
            assert (
                len(rv) <= 1
            ), f"Rank variable {rank_variable} indexes into multiple ranks {rv} for tensor {tensor} "
            return next(iter(rv), Rank("NO RANK. RECOMPUTED."))

        def make_loops(above_index: int, tensor_name: TensorName) -> list[MappingLoop]:
            loops = [
                n for n in mapping.nodes[:above_index] if isinstance(n, MappingLoop)
            ]
            loops = [
                Loop(
                    rank_name=get_rank(n.rank_variable, tensor_name),
                    tile_pattern=n.tile_pattern._symbol2str(),
                    is_spatial=isinstance(n, Spatial),
                )
                for n in loops
            ]
            return tuple(loops)

        return cls(
            tensors=fzs(
                TensorReservation(
                    name=mapping.nodes[i].purpose,
                    loops=make_loops(i, mapping.nodes[i].purpose),
                    resource_name=mapping.nodes[i].resource,
                    persistent=mapping.nodes[i].persistent,
                )
                for i in tensor_indices
            ),
            splits=fzs(
                Split(split=n, above_loop_index=i) for i in split_above_loop_indices
            ),
            reservation_indices=fzs(reservation_indices),
        )

    def symbols(self) -> list[str]:
        symbols = []

        def add(x: str):
            if isinstance(x, str) and x not in symbols:
                symbols.append(x)

        for t in self.tensors:
            for l in t.loops:
                add(l.tile_pattern.initial_tile_shape)
                add(l.tile_pattern.tile_shape)
                add(l.tile_pattern.calculated_n_iterations)
        return symbols

    def drop_loop_indices(self, loop_indices: set[int]) -> "Compatibility":
        loop_indices = set(loop_indices)
        tensors = fzs(t.drop_loop_indices(loop_indices) for t in self.tensors)
        splits = fzs(s for s in self.splits if s.above_loop_index not in loop_indices)

        def adjust(i: int) -> int:
            return i - sum(x < i for x in loop_indices)

        reservation_indices = fzs(adjust(i) for i in self.reservation_indices)
        reservation_indices = fzs(x for x in reservation_indices if x >= 0)

        splits = fzs(
            s.update(above_loop_index=adjust(s.above_loop_index)) for s in self.splits
        )
        return Compatibility(
            tensors=tensors,
            splits=splits,
            reservation_indices=reservation_indices,
        )

    def _prepend_symbols(self, prepend: str) -> "Compatibility":
        return self.update(
            tensors=fzs(t._prepend_symbols(prepend) for t in self.tensors)
        )

    def clear_tile_patterns_and_reservation_indices(self) -> "Compatibility":
        return self.update(
            tensors=fzs(t.clear_symbolic_tile_patterns() for t in self.tensors),
            reservation_indices=fzs(),
            check_reservation_indices=False,
        )

    def clear_symbolic_tile_patterns(self) -> "Compatibility":
        return self.update(
            tensors=fzs(t.clear_symbolic_tile_patterns() for t in self.tensors)
        )

    def make_fused_loop_symbols(
        self, prefix: str
    ) -> tuple[dict[str, str], "Compatibility"]:
        result = {}
        tensors = []
        for t in self.tensors:
            r, t = t.make_fused_loop_symbols(prefix)
            tensors.append(t)
            result.update(r)

        return result, self.update(tensors=fzs(tensors))

    def add_n_iteration_symbols(self) -> "Compatibility":
        return self.update(
            tensors=fzs(t.add_n_iteration_symbols() for t in self.tensors)
        )

    def clear_unrelated_columns(self, mappings: pd.DataFrame) -> "Compatibility":
        my_symbols = set(self.symbols())
        for c in my_symbols:
            assert c in mappings.columns, f"Column {c} not found in mappings"
        should_drop = lambda x: is_fused_loop_col(x) and x not in my_symbols
        drop = [c for c in mappings.columns if should_drop(c)]
        return mappings.drop(columns=drop)



from collections import defaultdict
from numbers import Number

from sympy import Symbol

from accelforge.frontend.workload import Workload
from accelforge.frontend._workload_isl._symbolic import get_stride_and_halo
from accelforge.frontend.mapping import (
    Loop,
    Mapping,
)


class SymbolRelations:
    def __init__(self):
        self.what_tiles_symbol: list[tuple[Symbol | int, Symbol | int]] = []
        self.tile_shape_and_initial: list[tuple[Symbol | int, Symbol | int]] = []
        self.delta_choices: list[tuple[Symbol, frozenset[int]]] = []
        self.bounds: tuple[tuple[Symbol, int, int], ...] = ()

    def make_bounds(self):
        all_symbols = set(
            s for w in self.what_tiles_symbol for s in w if isinstance(s, Symbol)
        )
        self.bounds = tuple((s, 1, self.get_max_size(s)) for s in all_symbols)

    def is_stride(self, symbol: Symbol) -> bool:
        """Check if `symbol` is a stride."""
        for tile_shape, initial in self.tile_shape_and_initial:
            if tile_shape == symbol:
                return True
            if initial == symbol:
                return False
        return True

    def is_initial_tile_shape(self, symbol: Symbol) -> bool:
        """Check if `symbol` is a initial tile shape."""
        for tile_shape, initial in self.tile_shape_and_initial:
            if tile_shape == symbol:
                return False
            if initial == symbol:
                return True
        return False

    def get_tile_shape(
        self, symbol: Symbol, none_if_fail: bool = False
    ) -> Symbol | int:
        """Get the stride corresponding to the initial tile shape `symbol`."""
        found = None
        for tile_shape, initial in self.tile_shape_and_initial:
            if initial == symbol:
                if found is None:
                    found = tile_shape
                elif found != tile_shape:
                    raise ValueError(
                        f"Symbol {symbol} has both {found} and {tile_shape} as initial"
                    )
        if found is not None or none_if_fail:
            return found
        raise ValueError(f"Symbol {symbol} not found as initial in {self}")

    def get_initial(self, symbol: Symbol, none_if_fail: bool = False) -> Symbol | int:
        found = None
        for tile_shape, initial in self.tile_shape_and_initial:
            if tile_shape == symbol:
                if found is None:
                    found = initial
                elif found != initial:
                    raise ValueError(
                        f"Symbol {symbol} has both {found} and {initial} as tile_shape"
                    )
        if found is not None or none_if_fail:
            return found
        raise ValueError(f"Symbol {symbol} not found as tile_shape in {self}")

    def get_delta_choices(self, symbol: Symbol) -> frozenset[int]:
        """Get the possible initial deltas for the rank variable represented by `symbol`."""
        for initial, choices in self.delta_choices:
            if initial == symbol:
                return choices
        raise ValueError(f"Symbol {symbol} not found in {self}")

    def get_inner_tiles(
        self, symbol: Symbol, none_if_fail: bool = False
    ) -> Symbol | int | None:
        """Get tiles within the tile represented by `symbol`."""
        found = None
        for tiled_by, what_tiles in self.what_tiles_symbol:
            if tiled_by == symbol:
                if found is None:
                    found = what_tiles
                elif found != what_tiles:
                    raise ValueError(
                        f"Symbol {symbol} is tiled by both {found} and {what_tiles}"
                    )
        if found is not None or none_if_fail:
            return found
        raise ValueError(f"Symbol {symbol} not found in {self}")

    def get_outer_tiles(
        self, symbol: Symbol, none_if_fail: bool = False
    ) -> Symbol | int | None:
        """Get the tile that contain the tile represented by `symbol`."""
        found = None
        for tiled_by, what_tiles in self.what_tiles_symbol:
            if what_tiles == symbol:
                if found is None:
                    found = tiled_by
                elif found != tiled_by:
                    raise ValueError(
                        f"Symbol {symbol} tiles both {tiled_by} and {what_tiles}"
                    )
        if found is not None or none_if_fail:
            return found
        raise ValueError(f"Symbol {symbol} not found in {self}")

    def get_max_size(self, symbol: Symbol) -> Number:
        while not isinstance(symbol, Number):
            symbol = self.get_outer_tiles(symbol)
        return symbol

    @staticmethod
    def from_pmapping_and_shape(
        pmapping: Mapping, shape: dict[str, int], workload: Workload
    ) -> "SymbolRelations":
        initial_delta_choices = get_initial_delta_choices(
            pmapping.nodes[-1].einsum, workload
        )

        relation = SymbolRelations()
        last_seen_loop_per_rank_var: dict[str, Symbol | int] = dict(shape)
        for node in pmapping.nodes:
            if not isinstance(node, Loop):
                continue
            prev = last_seen_loop_per_rank_var.get(node.rank_variable, None)
            # If we're a symbol and we've seen an outer loop with the same rank variable,
            # then we tile that one.
            if prev is not None:
                relation.what_tiles_symbol.append((prev, node.tile_shape))
            last_seen_loop_per_rank_var[node.rank_variable] = node.tile_shape

            if (
                isinstance(node.initial_tile_shape, Symbol)
                and node.initial_tile_shape != node.tile_shape
            ):
                relation.tile_shape_and_initial.append(
                    (node.tile_shape, node.initial_tile_shape)
                )
                relation.delta_choices.append(
                    (
                        node.initial_tile_shape,
                        frozenset(initial_delta_choices[node.rank_variable]),
                    )
                )

        for r, s in last_seen_loop_per_rank_var.items():
            if isinstance(s, Symbol):
                relation.what_tiles_symbol.append((s, 1))

        relation.make_bounds()
        return relation


def get_initial_delta_choices(einsum_name: str, workload: Workload):
    stride_and_halo = get_stride_and_halo(workload)
    einsum = workload.einsums[einsum_name]

    choices = defaultdict(lambda: set([0]))
    consumer_chains = []
    stack = [[(None, einsum)]]
    while stack:
        cur_chain = stack.pop()
        last_tensor, last_einsum = cur_chain[-1]
        for tensor in last_einsum.output_tensor_names:
            einsums_with_tensor_as_input = workload.einsums_with_tensor_as_input(tensor)

            if len(einsums_with_tensor_as_input) == 0:
                consumer_chains.append(cur_chain)

            for next_einsum in einsums_with_tensor_as_input:
                stack.append(cur_chain + [(tensor, next_einsum)])

    for chain in consumer_chains:
        for (_, producer), (tensor, consumer) in zip(
            list(reversed(chain))[1:], reversed(chain)
        ):
            rank_stride_and_halo = stride_and_halo[(consumer.name, tensor)]
            if tensor is None:
                break  # done

            for cons_rank_var in consumer.rank_variables:
                for prod_rank_var in producer.rank_variables:
                    prod_rank = prod_rank_var.upper()
                    for cons_choice in choices[cons_rank_var]:
                        key = (prod_rank, cons_rank_var)
                        if key not in rank_stride_and_halo:
                            continue
                        stride, halo = rank_stride_and_halo[key]
                        choices[prod_rank_var].add(int(cons_choice * stride + halo))

    return choices



# from bindings._looptree import LooptreeWorkload, LooptreeWorkloadDependencyAnalyzer


class EquivalentGroups:
    def __init__(self):
        self.group_id_to_ranks = {}
        self.rank_to_group_id = {}

    @staticmethod
    def from_workload(
        workload: "LooptreeWorkload", analyzer: "LooptreeWorkloadDependencyAnalyzer"
    ):
        einsum_id_to_name = workload.einsum_id_to_name()

        groups = EquivalentGroups()

        seen_ranks = set()
        for einsum_id in einsum_id_to_name:
            for rank_id in workload.einsum_ospace_dimensions(einsum_id):
                equiv_ranks = analyzer.equivalent_dimensions(einsum_id, rank_id)
                equiv_ranks = frozenset(equiv_ranks)
                if equiv_ranks not in seen_ranks:
                    seen_ranks.add(equiv_ranks)
                    group_id = len(groups.group_id_to_ranks)
                    groups.group_id_to_ranks[group_id] = equiv_ranks
                    for r in equiv_ranks:
                        groups.rank_to_group_id[r] = group_id

        return groups



import sympy
from sympy import Max
from sympy import Min

# class MaxGeqZero(Max):
#     def _eval_is_nonnegative(self, *args, **kwargs):
#         return True

#     def _eval_is_negative(self, *args, **kwargs):
#         return False

# class MinGeqZero(Min):
#     def _eval_is_nonnegative(self, *args, **kwargs):
#         return True

#     def _eval_is_negative(self, *args, **kwargs):
#         return False

# MaxGeqZero = Max
# MinGeqZero = Min


def MaxGeqZero(*args):
    non_max_args = [a for a in args if not isinstance(a, Max)]
    max_args = [a for a in args if isinstance(a, Max)]
    for a in max_args:
        non_max_args.extend(a.args)

    if 0 not in non_max_args:
        non_max_args.append(0)

    x = Max(*non_max_args)
    if isinstance(x, Max):
        x._eval_is_nonnegative = lambda *args, **kwargs: True
        x._eval_is_negative = lambda *args, **kwargs: False
    return x


def MinGeqZero(*args):
    non_min_args = [a for a in args if not isinstance(a, Min)]
    min_args = [a for a in args if isinstance(a, Min)]
    for a in min_args:
        non_min_args.extend(a.args)
    x = Min(*non_min_args)
    if isinstance(x, Min):
        x._eval_is_nonnegative = lambda *args, **kwargs: True
        x._eval_is_negative = lambda *args, **kwargs: False
    return x


# MAX BUG FIX.
# def Min(a, *bs):
#     """More post-lambdify broadcast-friendly option than sympy.Min"""
#     result = a
#     for b in bs:
#         result = sympy.Piecewise((result, result < b), (b, True))
#     return result

# def Max(a, *bs):
#     """More post-lambdify broadcast-friendly option than sympy.Max"""
#     result = a
#     for b in bs:
#         result = sympy.Piecewise((result, result > b), (b, True))
#     return result



from collections.abc import Mapping
import copy
from math import ceil, exp, prod
import random
import threading
import time
from accelforge._accelerated_imports import pd
from accelforge.mapper.simanneal.evalmapping import quick_join
from accelforge.mapper.simanneal.tracking import EvaluationsScoreTracker
from accelforge.mapper.FFM._join_pmappings.join_pmappings import PmappingGroup
from accelforge.mapper.FFM._join_pmappings.compatibility import (
    TensorReservation,
    Compatibility,
)
from accelforge.mapper.FFM._join_pmappings.pmapping_group import (
    MAPPING_COLUMN,
    PmappingDataframe,
)
from accelforge.util._frozenset import fzs
from accelforge.mapper.simanneal.mapspaceglobals import MapspaceGlobals

OBJECTIVE_COLUMN = None  # None -> Product


class FailedMutation(Exception):
    pass


class Mapping:
    def __init__(self, pmapping_groups: dict[str, list[PmappingGroup]]):
        self.einsum_names = list(pmapping_groups.keys())
        self.einsum2intra_choice = {
            einsum_name: None for einsum_name in self.einsum_names
        }
        self.einsum2tiling = {}
        for einsum_name, pm_group_list in pmapping_groups.items():
            tensor_names = pm_group_list[0].tensor_names
            tensors = fzs(TensorReservation(t, 0, 0, 0) for t in tensor_names)
            self.set_einsum2tiling(einsum_name, Compatibility(tuple(), tensors))

        # self.history = []
        class dummy_appender:
            def append(*args, **kwargs):
                pass

        self.history = dummy_appender()
        self.n_crossovers = 0
        self.n_mutations = 0

        self.n_changes = 0
        self.prev_eval_result = float("inf")
        self.prev_eval_at_n_changes = -1

    def set_einsum2tiling(self, einsum_name: str, tiling: Compatibility):
        prev = self.einsum2tiling.get(einsum_name, None)
        if prev is not None and prev == tiling:
            return
        self.einsum2tiling[einsum_name] = tiling
        self.einsum2intra_choice[einsum_name] = None

    def fix_loops(self, mapspace_globals: MapspaceGlobals):
        """Ensure that all tilings have the correct number of loops"""
        self.n_changes += 1
        self.history.append("Fixing loops")

        try:
            for einsum in self.einsum_names:
                tiling = self.einsum2tiling[einsum]
                n_loops = max(t.above_loop_index for t in tiling.tensors)

                # If there's too many loops then drop the extra ones
                if n_loops < len(tiling.loops):
                    self.set_einsum2tiling(
                        einsum, tiling.update(loops=tiling.loops[:n_loops])
                    )

                # If there's not enough loops then add some
                if n_loops > len(tiling.loops):
                    for tensor in tiling.tensors:
                        for loop in range(len(tiling.loops), tensor.above_loop_index):
                            self.mutate_loop(mapspace_globals, tensor, loop, einsum)
                            self.force_loop_match(mapspace_globals, loop, einsum)
                assert n_loops == len(self.einsum2tiling[einsum].loops)

                tiling = self.einsum2tiling[einsum]
                tensors = tiling.tensors
                for i in range(len(tiling.loops)):
                    tensors = list(t for t in tensors if t.above_loop_index > i)
                    if not tensors:
                        continue
                    possible_loops = set.intersection(
                        *(
                            mapspace_globals.tensor2possible_loops_above_set[einsum][t]
                            for t in tensors
                        )
                    )
                    if not possible_loops:
                        raise FailedMutation(
                            f"No possible loops above {i} for {einsum}"
                        )
                    if tiling.loops[i] not in possible_loops:
                        new_loop = random.choice(list(possible_loops))
                        self.history.append(
                            f"Fixing loop {i} for {einsum} to {new_loop}"
                        )
                        new_loops = (
                            tiling.loops[:i] + (new_loop,) + tiling.loops[i + 1 :]
                        )
                        self.set_einsum2tiling(einsum, tiling.update(loops=new_loops))

        except FailedMutation:
            self.history.append(f"Failed to fix loops")
            raise FailedMutation("Failed to fix loops")

    def match_loops(
        self, index: int, einsum_name: str, mapspace_globals: MapspaceGlobals
    ):
        """Ensure that loops match across Einsums"""
        self.n_changes += 1
        tiling = self.einsum2tiling[einsum_name]
        for einsum_name2, tiling2 in self.einsum2tiling.items():
            if einsum_name2 == einsum_name:
                continue
            shared_loop_index = max(
                tiling.shared_loop_index(tiling2.tensor_names),
                tiling2.shared_loop_index(tiling.tensor_names),
            )
            for i in range(min(shared_loop_index, index) + 1):
                # Translate loop from einsum_name to einsum_name2
                loop = tiling.loops[i]
                translations = mapspace_globals.rank_translations[einsum_name][
                    einsum_name2
                ][loop.rank_variable_name]
                if not translations:
                    raise FailedMutation(
                        f"Failed to translate loop {loop} from {einsum_name} to {einsum_name2}"
                    )
                rank_variable_name = random.choice(translations)
                new_loops = (
                    tiling2.loops[:i]
                    + (loop.update(rank_variable_names=fzs((rank_variable_name,))),)
                    + tiling2.loops[i + 1 :]
                )
                tiling2 = tiling2.update(loops=new_loops)
            self.set_einsum2tiling(einsum_name2, tiling2)

    def mutate_loop(
        self,
        mapspace_globals: MapspaceGlobals,
        tensor: TensorReservation = None,
        index: int = None,
        einsum_name: str = None,
    ):
        self.n_changes += 1
        if tensor is None:
            memories = set().union(*(t.tensors for t in self.einsum2tiling.values()))
            memories = [m for m in memories if m.above_loop_index > 0]
            if not memories:
                raise FailedMutation("No memories to mutate")
            tensor = random.choice(list(memories))
        if index is None:
            index = random.randint(0, tensor.above_loop_index - 1)
        if einsum_name is None:
            possible_einsums = [
                e for e, t in self.einsum2tiling.items() if tensor in t.tensors
            ]
            assert possible_einsums
            einsum_name = random.choice(possible_einsums)

        tiling = self.einsum2tiling[einsum_name]
        prev_loop = None

        choice = random.choice(["Increasing", "Decreasing", "Randomizing"])
        if len(tiling.loops) <= index:
            choice = "Randomizing"

        candidates = mapspace_globals.tensor2possible_loops_above[einsum_name][tensor]
        if choice == "Randomizing":
            new_loop = random.choice(candidates)
        else:
            prev_loop = tiling.loops[index]
            rank, bound = prev_loop.rank_variable_name, prev_loop.bound
            comparison = lambda x, y: x > y if choice == "Increasing" else x < y

            candidates = [
                c
                for c in candidates
                if comparison(c.bound, bound) and c.rank_variable_name == rank
            ]
            if not candidates:
                raise FailedMutation(
                    f"{choice} {prev_loop} for {einsum_name} at {index} failed"
                )
            new_loop = random.choice(candidates)

        self.history.append(f"{choice} loop {index} for {einsum_name} to {new_loop}")
        new_loops = tiling.loops[:index] + (new_loop,) + tiling.loops[index + 1 :]
        self.set_einsum2tiling(einsum_name, tiling.update(loops=new_loops))

    def get_shared_loop_index(
        self, mapspace_globals: MapspaceGlobals, einsum_name0: int, einsum_name1: int
    ):
        einsum_names = list(self.einsum2tiling.keys())
        if einsum_name0 == einsum_name1:
            einsum_name = einsum_names[einsum_index0]
            return len(self.einsum2tiling[einsum_name].loops) - 1

        einsum_index0 = einsum_names.index(einsum_name0)
        einsum_index1 = einsum_names.index(einsum_name1)

        if einsum_index0 > einsum_index1:
            einsum_index0, einsum_index1 = einsum_index1, einsum_index0

        tiling0 = self.einsum2tiling[einsum_names[einsum_index0]]
        tiling1 = self.einsum2tiling[einsum_names[einsum_index1]]
        left_tensors = mapspace_globals.get_tensors(*einsum_names[: einsum_index0 + 1])
        right_tensors = mapspace_globals.get_tensors(*einsum_names[einsum_index1:])
        return max(
            tiling0.shared_loop_index(right_tensors),
            tiling1.shared_loop_index(left_tensors),
        )

    def force_loop_match(
        self,
        mapspace_globals: MapspaceGlobals,
        index: int,
        einsum_name: str,
    ):
        self.n_changes += 1
        tiling = self.einsum2tiling[einsum_name]
        for einsum_name2, tiling2 in self.einsum2tiling.items():
            if einsum_name2 == einsum_name:
                continue
            shared_loop_index = self.get_shared_loop_index(
                mapspace_globals, einsum_name, einsum_name2
            )
            rank_translations = mapspace_globals.rank_translations[einsum_name][
                einsum_name2
            ]
            for i in range(min(shared_loop_index, index) + 1):
                loop = tiling.loops[i]
                translations = rank_translations[loop.rank_variable_name]
                if not translations:
                    raise FailedMutation(
                        f"Failed to translate loop {loop} from {einsum_name} to {einsum_name2}"
                    )
                rank_variable_name = random.choice(translations)
                new_loops = (
                    tiling2.loops[:i]
                    + (loop.update(rank_variable_names=fzs((rank_variable_name,))),)
                    + tiling2.loops[i + 1 :]
                )
                tiling2 = tiling2.update(loops=new_loops)
            self.set_einsum2tiling(einsum_name2, tiling2)

    def mutate_backing_tensor(self, mapspace_globals: MapspaceGlobals):
        self.n_changes += 1
        tensor = random.choice(
            list(mapspace_globals.tensor_names_used_in_multiple_einsums)
        )
        memories = random.choice(mapspace_globals.tensor2memories[tensor])
        for t in self.einsum2tiling.values():
            if memories in t.tensors:
                raise FailedMutation(
                    f"Moving tensor {tensor} to tensor {memories} failed"
                )
        self.history.append(f"Moving tensor {tensor} to tensor {memories}")
        for einsum, tiling in self.einsum2tiling.items():
            if not any(r.name == tensor for r in tiling.tensors):
                continue
            new_tensors = [memories] + [r for r in tiling.tensors if r.name != tensor]
            self.set_einsum2tiling(einsum, tiling.update(tensors=fzs(new_tensors)))
        self.fix_loops(mapspace_globals)

    def mutate_order(self, mapspace_globals: MapspaceGlobals):
        return
        self.n_changes += 1
        e0, e1 = random.sample(self.einsum_names, 2)
        print(f"Switching {e0} and {e1}")
        self.einsum2tiling[e0], self.einsum2tiling[e1] = (
            self.einsum2tiling[e1],
            self.einsum2tiling[e0],
        )
        self.fix_loops(mapspace_globals)

    def evaluate(self, mapspace_globals: MapspaceGlobals, return_df=False) -> float:
        if self.n_changes == self.prev_eval_at_n_changes and not return_df:
            return self.prev_eval_result, 1

        chosen_sims = []
        chosen_mappings = {}
        n_evaluations = (
            mapspace_globals.size_scale * mapspace_globals.find_pmapping_scale
        )

        if self.n_changes == self.prev_eval_at_n_changes and not return_df:
            return self.prev_eval_result, 1
        self.prev_eval_at_n_changes = self.n_changes
        self.prev_eval_result = float("inf")

        for einsum_name, t in self.einsum2tiling.items():
            if t not in mapspace_globals.einsum_tiling_2_sim[einsum_name]:
                assert not return_df
                return float("inf"), n_evaluations

            sim = mapspace_globals.einsum_tiling_2_sim[einsum_name][t]
            chosen_sims.append(sim)
            intra_mappings = sim.mappings.data

            if self.einsum2intra_choice[einsum_name] is not None:
                mapping = intra_mappings.iloc[
                    self.einsum2intra_choice[einsum_name] % len(intra_mappings)
                ]
                chosen_mappings[einsum_name] = mapping
                continue

            self.einsum2intra_choice[einsum_name] = random.randint(0, 1000000000000)
            choice = self.einsum2intra_choice[einsum_name] % len(sim.mappings.data)
            self.einsum2intra_choice[einsum_name] = choice
            n_evaluations += (
                mapspace_globals.size_scale * mapspace_globals.find_pmapping_scale
            )
            mapping = intra_mappings.iloc[choice]
            chosen_mappings[einsum_name] = mapping

        try:
            new_sims = {}
            for einsum_name, tiling in self.einsum2tiling.items():
                sim = mapspace_globals.einsum_tiling_2_sim[einsum_name][tiling]
                mapping_index = self.einsum2intra_choice[einsum_name] % len(
                    sim.mappings.data
                )
                new_sims[einsum_name] = [
                    PmappingGroup(
                        compatibility=sim.compatibility,
                        mappings=PmappingDataframe(
                            sim.mappings.data.iloc[
                                mapping_index : mapping_index + 1
                            ].copy()
                        ),
                    )
                ]
                chosen_mappings = quick_join(new_sims, mapspace_globals)
                assert len(chosen_mappings.data) == 1
                chosen_mappings = chosen_mappings.data.iloc[0]
        except Exception as e:
            assert not return_df
            return float("inf"), n_evaluations

        obj_cols = mapspace_globals.objective_function_cols
        score = prod(chosen_mappings[col] for col in obj_cols)
        # if score < 4.7770043942936216e+20:
        #     print("AHH")
        # import pydot
        # graph = pydot.Dot(graph_type="digraph", ranksep="0.2", nodesep="0.2")
        # tree.to_pydot(graph)
        # with open(f"test.png", "wb") as f:
        #     f.write(graph.create_png())

        if return_df:
            d = {col: sum(c[col] for c in chosen_mappings.values()) for col in obj_cols}
            d[MAPPING_COLUMN] = mapping
            self.prev_eval_result = score
            return pd.DataFrame([d]), n_evaluations
        self.prev_eval_result = score
        return score, n_evaluations

    def mutate_intra_mapping(self, mapspace_globals: MapspaceGlobals):
        self.n_changes += 1
        einsum_name = random.choice(self.einsum_names)
        self.history.append(f"Choosing intra-layer mapping for {einsum_name}")
        self.einsum2intra_choice[einsum_name] = None

    def get_mutation_functions(self):
        return [
            self.mutate_loop,
            self.mutate_backing_tensor,
            self.mutate_order,
            self.mutate_intra_mapping,
        ]

    def crossover(self, other: Mapping, mapspace_globals: MapspaceGlobals):
        child = copy.deepcopy(other)
        einsum_name = random.choice(child.einsum_names)
        try:
            child.set_einsum2tiling(einsum_name, self.einsum2tiling[einsum_name])
            child.einsum2intra_choice[einsum_name] = self.einsum2intra_choice[
                einsum_name
            ]
            child.n_changes += 1
            for i in range(len(child.einsum2tiling[einsum_name].loops)):
                child.match_loops(i, einsum_name, mapspace_globals)
            child.fix_loops(mapspace_globals)
            child.n_crossovers += 1
        except FailedMutation:
            return copy.deepcopy(other)
        return child

    @staticmethod
    def create_random_mapping(mapspace_globals: MapspaceGlobals):
        mapping = Mapping(mapspace_globals.pmapping_groups)
        prev_compatibility: Compatibility = None
        einsum_names = list(mapping.einsum2tiling.keys())
        for i, einsum_name in enumerate(einsum_names):
            pm_group_list = mapspace_globals.pmapping_groups[einsum_name]
            if prev_compatibility is None:
                sim = random.choice(pm_group_list)
                mapping.set_einsum2tiling(einsum_name, sim.compatibility)
                if len(einsum_names) == 1:
                    break
                prev_compatibility = mapspace_globals.compatibility2rightcompatibility[
                    einsum_name
                ][sim.compatibility]
                live_tensors = mapspace_globals.get_live_tensors(*einsum_names[i + 1 :])
                prev_compatibility = prev_compatibility.clear_dead_tensors(
                    live_tensors=live_tensors
                )
                continue

            tilings = []
            compatiblity_options = mapspace_globals.leftcompatibility2tiling[
                einsum_name
            ]
            cur_tensors = mapspace_globals.get_tensors(einsum_name)
            for translation in mapspace_globals.get_possible_translations(
                prev_compatibility, einsum_name
            ):
                translation = translation.clear_dead_tensors(
                    live_tensors=cur_tensors, keep_loops=True
                )
                if translation in compatiblity_options:
                    tilings.extend(compatiblity_options[translation])

            if not tilings:
                raise FailedMutation(
                    f"No tilings for {einsum_name} with {prev_compatibility}"
                )
            sim_choices = [
                mapspace_globals.einsum_tiling_2_sim[einsum_name][t] for t in tilings
            ]
            sim = random.choice(sim_choices)
            tiling = sim.compatibility
            mapping.set_einsum2tiling(einsum_name, tiling)
            if i == len(einsum_names) - 1:
                break
            new_compatibility: Compatibility = (
                mapspace_globals.compatibility2rightcompatibility[einsum_name][tiling]
            )
            # Combine prev_compatibility and new_compatibility
            live_tensors = mapspace_globals.get_live_tensors(*einsum_names[i + 1 :])
            prev_compatibility = prev_compatibility.merge_next(
                new_compatibility, live_tensors
            )
        return mapping


def get_accept_function(temperature, cooling_rate, evaluations_tracker):
    proportion = evaluations_tracker.evaluations / evaluations_tracker.max_evaluations
    new_temp = temperature * (1 - proportion) / (1 + cooling_rate * proportion)

    # Assume prescient knowledge of the best score with which to scale by
    def accept(prev_eval_result, new_score):
        if new_score == float("inf"):
            return False
        if new_score <= prev_eval_result:
            return True
        scaleby = new_temp * evaluations_tracker.stop_at_score
        if scaleby > 0 and random.random() < exp(
            (prev_eval_result - new_score) / scaleby
        ):
            return True
        return False

    return accept


def mutate(
    mapping: Mapping, mapspace_globals: MapspaceGlobals, accept_function: callable
):
    prev_mapping = copy.deepcopy(mapping)
    prev_eval_result = mapping.prev_eval_result
    n_evaluations = 1
    try:
        choice = random.choice(mapping.get_mutation_functions())
        choice(mapspace_globals)
    except FailedMutation:
        return prev_mapping, n_evaluations
    new_score, n_evaluations = mapping.evaluate(mapspace_globals)
    if new_score == float("inf"):
        return prev_mapping, n_evaluations
    if accept_function(prev_eval_result, new_score):
        return mapping, n_evaluations
    return prev_mapping, n_evaluations


def _fuse_sims(
    mapspace_globals: MapspaceGlobals,
    n_threads: int,
    evaluations_tracker: EvaluationsScoreTracker,
    algorithm: str,
):
    random.seed(time.time() + hash(threading.get_ident()))  # Seed with thread ID
    evaluations_tracker.multiply_scale_by(len(mapspace_globals.einsum_names))
    evaluations_tracker.print_period *= n_threads
    evaluations_tracker.max_evaluations //= n_threads

    def anneal_population(population, mapspace_globals: MapspaceGlobals, n_rounds):
        temperature = 0.07
        cooling_rate = 8
        while True:
            accept_function = get_accept_function(
                temperature, cooling_rate, evaluations_tracker
            )
            # population = parallel([delayed(mutate)(m, mapspace_globals, accept_function) for m in population])
            for j, mapping in enumerate(population):
                population[j], evaluations = mutate(
                    mapping, mapspace_globals, accept_function
                )
                if evaluations_tracker.add_evaluation(
                    evaluations, population[j].prev_eval_result
                ):
                    return population

    def genetic_algorithm_population(
        population, mapspace_globals: MapspaceGlobals, n_rounds
    ):
        population_size = len(population)
        crossover_rate = 0.7
        mutation_rate = 0.2

        def crossover(parent1: Mapping, parent2: Mapping):
            if random.random() > crossover_rate:
                return copy.deepcopy(parent1)
            return parent1.crossover(parent2, mapspace_globals)

        def mutate_individual(individual):
            individual = copy.deepcopy(individual)
            prev_mapping = copy.deepcopy(individual)
            if random.random() > mutation_rate:
                return individual
            try:
                mutation_function = random.choice(individual.get_mutation_functions())
                mutation_function(mapspace_globals)
                individual.n_mutations += 1
                return individual
            except FailedMutation:
                return prev_mapping

        best_fitness = float("inf")
        while True:
            # Evaluate fitness
            fitness = [0] * len(population)
            for i, individual in enumerate(population):
                f, evaluations = individual.evaluate(mapspace_globals)
                fitness[i] = f
                best_fitness = min(best_fitness, f)
                if evaluations_tracker.add_evaluation(evaluations, best_fitness):
                    return population

            best_score = min(fitness)
            best_mapping = population[fitness.index(best_score)]

            # Selection (roulette wheel selection)
            total_fitness = sum(1.0 / (f + 1e-9) for f in fitness)
            probabilities = [(1.0 / (f + 1e-9)) / total_fitness for f in fitness]
            selected_indices = random.choices(
                range(len(population)), probabilities, k=population_size
            )

            # Crossover
            new_population = list(population[i] for i in selected_indices)
            for i in range(0, population_size, 2):
                parent1 = population[selected_indices[i]]
                parent2 = population[selected_indices[(i + 1) % population_size]]
                child1 = crossover(parent1, parent2)
                child2 = crossover(parent2, parent1)
                new_population.extend([child1, child2])

            # Mutation
            for i, individual in enumerate(new_population):
                new_population[i] = mutate_individual(individual)

            new_population.append(best_mapping)  # Keep the best mapping around
            population = new_population

        return population

    def random_sample_population(
        population, mapspace_globals: MapspaceGlobals, n_rounds, prune=False
    ):
        best_mapping = population[0]
        best_score = float("inf")
        while True:
            try:
                mapping = Mapping.create_random_mapping(mapspace_globals)
            except FailedMutation:
                if not prune:
                    if evaluations_tracker.add_evaluation(1, float("inf")):
                        return [best_mapping]
                continue
            score, evaluations = mapping.evaluate(mapspace_globals)
            if score < best_score:
                best_mapping = mapping
                best_score = score
            if evaluations_tracker.add_evaluation(evaluations, score):
                return [best_mapping]
        return [best_mapping]

    extra_args = {}
    if algorithm == "genetic":
        population_size = 1000
        callfunc = genetic_algorithm_population
    elif algorithm == "simulated_anneal":
        population_size = 100 // n_threads
        callfunc = anneal_population
    elif "random" in algorithm:
        population_size = 1
        callfunc = random_sample_population
        extra_args["prune"] = "pruned" in algorithm

    # Randomly intialize the population
    def get_random_mapping():
        while True:
            try:
                mapping = Mapping.create_random_mapping(mapspace_globals)
                score, evaluations = mapping.evaluate(mapspace_globals)
                evaluations_tracker.add_evaluation(evaluations, score)
                if score == float("inf"):
                    raise FailedMutation("Random mapping failed")
                return mapping
            except FailedMutation:
                pass

    population = []
    while len(population) < population_size:
        try:
            mapping = Mapping.create_random_mapping(mapspace_globals)
            score, evaluations = mapping.evaluate(mapspace_globals)
            if evaluations_tracker.add_evaluation(evaluations, score):
                break
            if score == float("inf"):
                raise FailedMutation("Random mapping failed")
            population.append(mapping)
        except FailedMutation:
            if evaluations_tracker.add_evaluation(1, float("inf")):
                break

    n_rounds = 9999999999999999999999999
    results = callfunc(population, mapspace_globals, n_rounds)
    eval_results = []
    for m in results:
        try:
            eval_results.append(m.evaluate(mapspace_globals, return_df=True)[0])
        except:
            pass
    try:
        return pd.DataFrame(), evaluations_tracker
        assert False, "Not saving chosen mappings to avoid big files"
        return (
            pd.concat(eval_results),
            evaluations_tracker,
        )  # <- Resulted in large files bc it's not pareto pruned
    except Exception as e:
        for i in range(30):
            print(f"Failed to concatenate results. Exception: {e}")
        return pd.DataFrame(), evaluations_tracker



from sympy import Symbol
import accelforge.frontend.arch as arch
from accelforge.frontend.mapping import TensorHolder
from accelforge.mapper.FFM._make_pmappings.pmapper_job import Job
from accelforge.model._looptree.reuse import symbolic
from accelforge.model._looptree.reuse.symbolic import (
    analyze_reuse_and_add_reservations_to_mapping,
)
from accelforge.model._looptree.energy import (
    compute_energy_from_actions,
    gather_actions,
)
from accelforge.model._looptree.latency.memory import component_latency
from accelforge.mapper.FFM._join_pmappings.pmapping_dataframe import (
    nameloop2col,
    tensor2col,
    firstlatency2col,
    action2col,
    energy2col,
)
from accelforge.frontend.mapper.metrics import Metrics
import sympy
from numbers import Number
from accelforge.util._sympy.broadcast_max import MaxGeqZero


def run_model(
    job: Job,
    add_reservations: bool = True,
) -> tuple[list[Symbol], dict[str, float], dict[str, float], dict[str, float]]:
    pmapping = job.mapping
    spec = job.spec
    metrics = job.metrics
    is_copy_op = job.is_copy_operation
    workload = spec.workload

    df = {}

    reuse = analyze_reuse_and_add_reservations_to_mapping(
        job, add_reservations=add_reservations
    )

    latency = component_latency(reuse, job.flattened_arch, pmapping, spec)
    try:
        overall_latency = MaxGeqZero(*latency.values())
    except Exception as e:
        for k, v in latency.items():
            if not isinstance(v, (Number, sympy.Symbol, sympy.Expr)):
                raise ValueError(
                    f"Invalid type for latency: {k}: {type(v)} {str(v).strip()}"
                )

        raise ValueError(
            f"Error calculating latency for {job.einsum_name}. Could not calculate "
            f"a symbolic max of the following latencies:\n\t"
            + "\n\t".join(
                [f"{k}: {type(v)} {str(v).strip()}" for k, v in latency.items()]
            )
        )

    used_fanout = {
        (component, dim): n
        for (component, einsum), dims in reuse.fanout.items()
        if einsum == job.einsum_name
        for dim, n in dims.items()
    }
    # If there's no loops that use this spatial fanout, the model won't
    # output any usage. We still reserve at least one spatial instance.
    for f in job.flattened_arch:
        if isinstance(f, arch.Spatialable):
            for s in f.spatial:
                used_fanout.setdefault((f.name, s.name), 1)

    # Scale used fanout to get actual usage
    spatial_usage = {}
    spatial_usage_df = {}
    memory_to_size = {}
    for node in job.flattened_arch:
        if isinstance(node, arch.Memory):
            memory_to_size[node.name] = node.size

        if isinstance(node, arch.Spatialable):
            for s in node.spatial:
                usage = used_fanout[node.name, s.name] / s.fanout
                scaled_usage = usage * s.usage_scale
                spatial_usage[node.name, s.name] = scaled_usage
                s = f"usage<SEP>spatial<SEP>{node.name}<SEP>{s.name}"
                spatial_usage_df[s] = scaled_usage

    component_to_non_power_gated_porp, _ = spec.arch._power_gating(
        compute_name=job.flattened_arch[-1].name,
        used_fanout=spatial_usage,
    )

    if metrics & Metrics.ACTIONS:
        df.update(spatial_usage_df)

    actions = gather_actions(reuse, None, use_name=True)
    energy = compute_energy_from_actions(
        spec, actions, overall_latency, component_to_non_power_gated_porp
    )

    fusable_tensors = workload.tensor_names_used_in_multiple_einsums
    tensor_to_backing = {}
    for node in pmapping.nodes:
        if isinstance(node, TensorHolder):
            for tensor in node.tensors:
                if tensor not in tensor_to_backing and tensor in fusable_tensors:
                    tensor_to_backing[tensor] = node.component

    total_occupancy = {}
    compute_unit = pmapping.nodes[-1].component

    n_instances = workload.n_instances * workload.einsums[job.einsum_name].n_instances

    n_loop_options = set()
    for buffet, stats in reuse.buffet_stats.items():
        if buffet.level == compute_unit:
            continue

        occupancy = stats.max_occupancy

        if occupancy == 0:
            continue
        if stats.persistent:
            occupancy *= n_instances

        for tensor, backing in tensor_to_backing.items():
            if (is_copy_op or buffet.tensor == tensor) and buffet.level == backing:
                df[tensor2col(tensor)] = occupancy / memory_to_size[buffet.level]

        total_occupancy.setdefault(buffet.level, {}).setdefault(stats.n_loops_above, 0)
        total_occupancy[buffet.level][stats.n_loops_above] += occupancy
        n_loop_options.add(stats.n_loops_above)

    for memory, occupancies in total_occupancy.items():
        if memory not in job.memories_track_all:
            continue
        running_total = 0
        for n_loop in sorted(n_loop_options):
            if n_loop in occupancies:
                running_total += occupancies[n_loop]
                df[nameloop2col(memory, n_loop)] = (
                    running_total / memory_to_size[memory]
                )

    if metrics & Metrics.ACTIONS:
        detailed_actions = gather_actions(reuse, None, verbose=True, use_name=True)
        for key, count in detailed_actions.items():
            df[action2col(key)] = count.total * n_instances
        detailed_energy = compute_energy_from_actions(
            spec, detailed_actions, overall_latency, component_to_non_power_gated_porp
        )
        for key, energy_val in detailed_energy.items():
            df[energy2col(key)] = energy_val * n_instances
        for component, cur_latency in latency.items():
            df[f"latency<SEP>{component}"] = cur_latency * n_instances

    if metrics & Metrics.LATENCY:
        df["Total<SEP>latency"] = overall_latency * n_instances
        # df[f"latency<SEP>compute"] = comp_latency * n_instances
        # For first latency, we'll follow the convention of treating compute
        # as a component, similarly to memory (see below).
        for compute_level, stats in reuse.compute_stats.items():  # FIRST LATENCY
            for idx, max_first_latency in stats.max_first_latency.items():
                df[firstlatency2col(compute_level.level, idx)] = (
                    max_first_latency * n_instances
                )

    if metrics.includes_dynamic_energy():
        dynamic_energy = [e for k, e in energy.items() if k.action != "leak"]
        df["Total<SEP>dynamic_energy"] = sum(dynamic_energy) * n_instances

    if metrics.includes_leak_energy():
        leak_energy = [e for k, e in energy.items() if k.action == "leak"]
        df["Total<SEP>leak_energy"] = sum(leak_energy) * n_instances

    per_memory_spatial_usage_df = {}
    for memory, occupancies in total_occupancy.items():
        ignored = job.ignored_resources is not None and memory in job.ignored_resources
        key = f"usage<SEP>memory<SEP>{memory}"
        if not ignored:
            per_memory_spatial_usage_df[key] = (
                sum(occupancies.values()) / memory_to_size[memory]
            )
        if metrics & Metrics.ACTIONS:
            df[key] = sum(occupancies.values()) / memory_to_size[memory]

    if symbolic.PRINT_FORMULAS:
        for k, v in energy.items():
            print(f"{k}: {v}")
        for k, v in spatial_usage_df.items():
            print(f"{k}: {v}")
        for k, v in df.items():
            print(f"{k}: {v}")

    return (
        reuse.symbols,
        df,
        per_memory_spatial_usage_df,
        spatial_usage_df,
        reuse.tensor2mapping,
    )



from unittest import TestCase

import accelforge as af


class TestParsing(TestCase):
    def test_hierarchical(self):
        spec = af.Spec.from_yaml(
            af.examples.arches.networked.hierarchical,
        )
        self.assertIn("PeArray", spec.arch.nodes)
        self.assertEqual(spec.arch.nodes["PeArray"].get_fanout(), 1)
        self.assertIn("Scratchpad", spec.arch.nodes)
        self.assertEqual(spec.arch.nodes["Scratchpad"].get_fanout(), 4)
        self.assertIn("MacArray", spec.arch.nodes)
        self.assertEqual(spec.arch.nodes["MacArray"].get_fanout(), 1)

        try:
            spec = spec.calculate_component_area_energy_latency_leak()
        except af.EvaluationError as e:
            self.fail(e.message)

    def test_flat(self):
        spec = af.Spec.from_yaml(
            af.examples.arches.networked.flat,
        )
        self.assertIn("NoC", spec.arch.nodes)
        self.assertEqual(spec.arch.nodes["NoC"].get_fanout(), 1)
        self.assertEqual(
            {n.name for n in spec.arch.get_nodes_of_type(af.spec.Leaf)},
            {'MainMemory', 'GlobalBuffer', 'NoC', 'RowBuffer', 'ColumnBuffer', 'DistributedBuffer', 'Scratchpad', 'MAC'}
        )

        try:
            spec = spec.calculate_component_area_energy_latency_leak()
        except af.EvaluationError as e:
            self.fail(e.message)


class TestModel(TestCase):
    def test_hierarchical(self):
        M = 8
        KN = 8
        MAC_TILE = 2
        PE_TILE = KN//MAC_TILE
        M_TILE = 4
        BITS_PER_VALUE = 8

        spec = af.Spec.from_yaml(
            af.examples.workloads.matmuls,
            af.examples.arches.networked.hierarchical,
            af.examples.mappings.one_matmul_to_networked_hierarchical,
            jinja_parse_data={"N_EINSUMS": 1, "M": 8, "KN": 8, "MAC_TILE": MAC_TILE, "M_TILE": M_TILE}
        )
        result = spec.evaluate_mapping()
        self.assertEqual(
            result.data["Matmul0<SEP>action<SEP>MacArray<SEP>T0<SEP>hops"].iloc[0],
            (M/M_TILE)*(KN/MAC_TILE)**2 * M_TILE * (0.5*MAC_TILE*(MAC_TILE-1) + MAC_TILE*(MAC_TILE-1)) * BITS_PER_VALUE
        )
        # NOTE: assuming XY routing (as defined in mapping)
        self.assertEqual(
            result.data["Matmul0<SEP>action<SEP>MacArray<SEP>T1<SEP>hops"].iloc[0],
            (M/M_TILE)*(KN/MAC_TILE)**2 * M_TILE * (MAC_TILE*(MAC_TILE-1) + MAC_TILE*(MAC_TILE-1)) * BITS_PER_VALUE
        )
        self.assertEqual(
            result.data["Matmul0<SEP>action<SEP>MacArray<SEP>W0<SEP>hops"].iloc[0],
            (M/M_TILE)*(KN/MAC_TILE)**2 * M_TILE * (MAC_TILE*(MAC_TILE-1) + MAC_TILE*(MAC_TILE-1)) * BITS_PER_VALUE
        )

        self.assertEqual(
            result.data["Matmul0<SEP>action<SEP>PeArray<SEP>T0<SEP>hops"].iloc[0],
            (M/M_TILE) * (0.5*PE_TILE*(PE_TILE-1) + PE_TILE*(PE_TILE-1)) * M_TILE*MAC_TILE*BITS_PER_VALUE
        )
        # NOTE: assuming XY routing (as defined in mapping)
        self.assertEqual(
            result.data["Matmul0<SEP>action<SEP>PeArray<SEP>T1<SEP>hops"].iloc[0],
            (M/M_TILE) * (PE_TILE*(PE_TILE-1) + PE_TILE*0.5*PE_TILE*(PE_TILE-1)) * M_TILE*MAC_TILE*BITS_PER_VALUE
        )
        self.assertEqual(
            result.data["Matmul0<SEP>action<SEP>PeArray<SEP>W0<SEP>hops"].iloc[0],
            (M/M_TILE) * (PE_TILE*0.5*PE_TILE*(PE_TILE-1) + PE_TILE*0.5*PE_TILE*(PE_TILE-1)) * MAC_TILE**2*BITS_PER_VALUE
        )
